#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright 2018 Spanish National Research Council (CSIC)
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import os
import sys

from aiohttp import web

import deep_dashboard
from deep_dashboard import config
from deep_dashboard import log


CONF = config.CONF

LOG = log.LOG

INTRO = """
         ##         ###
         ##       ######  ##
     .#####   #####   #######.  .#####.
    ##   ## ## //   ##  //  ##  ##   ##
    ##. .##  ###  ###   // ###  ##   ##
      ## ##    ####     ####    #####.
              Hybrid-DataCloud  ##
"""


def main():
    print(INTRO)

    config.configure(sys.argv[1:])

    LOG.info("Starting DEEP dashboard version %s", deep_dashboard.__version__)

    app = deep_dashboard.init(sys.argv[1:])

    try:
        web.run_app(
            app,
            handle_signals=True,
            host=CONF.listen_ip,
            port=CONF.listen_port,
            path=CONF.listen_path
        )
    finally:
        if CONF.listen_path:
            os.remove(CONF.listen_path)


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-

# Copyright 2019 Spanish National Research Council (CSIC)
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import copy
import hashlib
import hmac

from aiohttp import web
import aiohttp_jinja2
import aiohttp_session_flash as flash
import markdown

from deep_dashboard import config
from deep_dashboard import deep_oc
from deep_dashboard import log

CONF = config.CONF
LOG = log.LOG

routes = web.RouteTableDef()


@routes.get('/modules', name="modules")
@aiohttp_jinja2.template('modules/index.html')
async def index(request):
    request.context["templates"] = await request.app.cache.modules.get_all()
    request.context["breadcrumbs"] = [
        ("Home", False, "/"),
        ("Modules", True, "/modules"),  # FIXME(aloga): use url
    ]
    return request.context


@routes.post("/reload", name="reload")
async def reload_all_modules(request):
    """Load TOSCA templates and map them to modules

    This function is used to refresh the TOSCA templates and the mapping
    between modules and TOSCA templates.  A webhook is set up so that when any
    of the repos [1][2] is updated, Github will POST to this method to refresh
    the Dashboard. The webhook's secret has to be the same has GITHUB_SECRET in
    the conf so that we can validate that the payload comes indeed from Github
    and the webhook has to be configured to deliver an 'application/json'.

    [1] https://github.com/deephdc/deep-oc
    [2] https://github.com/indigo-dc/tosca-templates/tree/master/deep-oc
    [3] https://gist.github.com/categulario/deeb41c402c800d1f6e6
    """

    # Check request comes indeed from Github
    if CONF.github_secret:
        if 'X-Hub-Signature' not in request.headers:
            return web.Response(
                text='Refresh petitions must be signed from Github.',
                status=403
            )
        # FIXME(aloga): this does not work
        signature = hmac.new(
            CONF.github_secret,
            request.data,
            hashlib.sha1
        ).hexdigest()
        if not hmac.compare_digest(
            signature,
            request.headers['X-Hub-Signature'].split('=')[1]
        ):
            return web.Response(
                text='Failed to verify the signature!',
                status=403
            )

    LOG.info('Reloading modules and TOSCA templates ...')
    await deep_oc.download_catalog(request.app)
    await deep_oc.load_catalog(request.app)

    return web.Response(status=201)


@routes.get("/modules/{module}/train", name="module.train")
@aiohttp_jinja2.template('modules/train.html')
async def configure_module_training(request):
    module = request.match_info["module"].lower()

    if not await request.app.cache.modules.exists(module):
        flash.flash(
            request,
            ("danger", f"Module does not exist: {module}.")
        )
        return web.HTTPFound("/modules")

    request.context["selected_module"] = module
    module_meta = await request.app.cache.modules.get(module)

    selected_tosca = request.query.get(
        "selected_tosca",
        list(module_meta["tosca_templates"].keys())[0]
    )
    template_name = module_meta["tosca_templates"][selected_tosca]
    hardware_configuration = request.query.get("hardware_configuration",
                                               "CPU").lower()
    docker_tag = request.query.get("docker_tag",
                                   module_meta["docker_tags"][0]).lower()
    run_command = request.query.get("run_command",
                                    "DEEPaaS")

    general_configuration = {
        "tosca_templates": {
            "available": module_meta["tosca_templates"].keys(),
            "selected": selected_tosca,
        },
        "docker_tags": {
            "available": module_meta["docker_tags"],
            "selected": docker_tag,
        },
        "hardware_configuration": {
            "available": ["CPU", "GPU"],
            "selected": hardware_configuration,
        },
        "run_command": {
            "available": ["DEEPaaS", "JupyterLab", "Custom command"],
            "selected": run_command,
        },
    }

    tosca_template = module_meta["tosca_templates"].get(selected_tosca)
    if tosca_template is None:
        flash.flash(
            request,
            ("danger", f"TOSCA template does not exist: {tosca_template}.")
        )
        return web.HTTPFound("/modules")

    aux = await request.app.cache.tosca_templates.get(tosca_template)
    inputs = copy.deepcopy(
        aux["inputs"]
    )
    inputs['docker_image'].setdefault(
        'default',
        module_meta['sources']['docker_registry_repo'])

    docker_tags = module_meta['docker_tags']
    if docker_tag not in docker_tags:
        docker_tag = docker_tags[0]

    if run_command == 'deepaas':
        inputs['run_command']['default'] = 'deepaas-run --listen-ip=0.0.0.0'
        if hardware_configuration == 'gpu':
            inputs['run_command']['default'] += ' --listen-port=$PORT0'
    elif run_command == 'jupyterlab':
        flash.flash(
            request,
            ("warning", 'Remember to set a Jupyter password (mandatory).')
        )
        inputs['run_command']['default'] = (
            '/srv/.deep-start/run_jupyter.sh --allow-root'
        )
        if hardware_configuration == 'gpu':
            inputs['run_command']['default'] = (
                "jupyterPORT=$PORT2 " + inputs['run_command']['default']
            )

    if hardware_configuration == 'cpu':
        inputs['num_cpus']['default'] = 1
        inputs['num_gpus']['default'] = 0
        inputs['run_command']['default'] = (
            "monitorPORT=6006 " + inputs['run_command']['default']
        )
    elif hardware_configuration == 'gpu':
        inputs['num_cpus']['default'] = 1
        inputs['num_gpus']['default'] = 1
        inputs['run_command']['default'] = (
            "monitorPORT=$PORT1 " + inputs['run_command']['default']
        )

    # FIXME(aloga): improve conditions here
    if run_command == "custom command":
        inputs['run_command']['default'] = ''

    inputs['docker_image']['default'] += ':{}'.format(docker_tag)

    # Group tosca options
    grouped = {
        "docker": {},
        "jupyter": {},
        "storage": {},
        "hardware": {},
        "other": {},
    }

    for k, v in inputs.items():
        if k.startswith("docker_"):
            grouped["docker"][k] = v
        elif k.startswith("jupyter_"):
            grouped["jupyter"][k] = v
        elif any([k.startswith("rclone_"),
                  k.startswith("onedata_"),
                  k.startswith("oneclient_"),
                  k == "app_in_out_base_dir"]):
            grouped["storage"][k] = v
        elif k in ["mem_size", "num_cpus", "num_gpus"]:
            grouped["hardware"][k] = v
        else:
            grouped["other"][k] = v

    # Remove empty groups
    for k, v in list(grouped.items()):
        if not v:
            del grouped[k]

    template_meta = {
        "inputs": inputs,
        "grouped": grouped,
    }

    request.context["general_configuration"] = general_configuration
    request.context["template_meta"] = template_meta
    request.context["template_name"] = template_name
    request.context["slas"] = request.app.slas
    request.context["module_meta"] = module_meta
    request.context["breadcrumbs"] = [
        ("Home", False, "/"),
        ("Modules", False, "/modules"),  # FIXME(aloga): use url
        (module, False, f"/modules/{module}"),  # FIXME(aloga): use url
        ("train", True, f"/modules/{module}/train"),  # FIXME(aloga): use url
    ]

    return request.context


@routes.get("/modules/{module}", name="module")
@aiohttp_jinja2.template('modules/module.html')
async def module_info(request):
    module = request.match_info["module"].lower()

    if not await request.app.cache.modules.exists(module):
        flash.flash(
            request,
            ("danger", f"Module does not exist: {module}.")
        )
        return web.HTTPFound("/modules")

    module_meta = await request.app.cache.modules.get(module)

    request.context["modulename"] = module
    request.context["module_meta"] = copy.deepcopy(module_meta)
    description = module_meta.get("description")
    if description:
        description = "\n".join(description)
    else:
        description = "No description provided."

    description = markdown.markdown(description)
    request.context["module_meta"]["description"] = description

    request.context["breadcrumbs"] = [
        ("Home", False, "/"),
        ("Modules", False, "/modules"),  # FIXME(aloga): use url
        (module, True, f"/modules/{module}"),  # FIXME(aloga): use url
    ]

    return request.context



# -*- coding: utf-8 -*-

# Copyright 2019 Spanish National Research Council (CSIC)
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from aiohttp import web
import aiohttp_jinja2
import aiohttp_security
import aiohttp_session
import aiohttp_session.cookie_storage

from deep_dashboard import auth
from deep_dashboard import sla
from deep_dashboard import utils

routes = web.RouteTableDef()


@routes.get('/', name="home")
@aiohttp_jinja2.template('home.html')
async def home(request):
    session = await aiohttp_session.get_session(request)

    next_url = session.get("next")
    if next_url and request.context["current_user"]["authenticated"]:
        del session["next"]
        return web.HTTPFound(next_url)

    request.context["breadcrumbs"] = [
        ("Home", True, "/"),
    ]

    return request.context


@routes.get('/logout', name='logout')
async def logout(request):
    redirect_response = web.HTTPFound('/')
    await aiohttp_security.forget(request, redirect_response)
    return redirect_response


@routes.get('/login/iam', name="login")
async def iam_login(request):
    client = request.app.iam_client

    # Check if is not redirect from provider
    if client.shared_key not in request.query:
        # Redirect client to provider
        return web.HTTPTemporaryRedirect(
            client.get_authorize_url(access_type='offline')
        )

    try:
        meta, userinfo = await auth.get_token_userinfo(request)
    except web.HTTPInternalServerError as e:
        raise e
    except Exception:
        return web.HTTPTemporaryRedirect(
            client.get_authorize_url(access_type='offline')
        )

    session = await aiohttp_session.get_session(request)
    session["userinfo"] = userinfo
    session["username"] = userinfo["name"]
    session["gravatar"] = utils.avatar(userinfo["email"], 26)
    session['organisation_name'] = userinfo['organisation_name']

    user_id = userinfo["sub"]

    redirect_response = web.HTTPFound("/")
    await aiohttp_security.remember(request, redirect_response, user_id)

    request.app.sla_loader = await sla.load_slas_as_task(request)

    return redirect_response



# -*- coding: utf-8 -*-

# Copyright 2019 Spanish National Research Council (CSIC)
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import datetime

import aioauth_client
from aiohttp import web
from aiohttp import web_urldispatcher
import aiohttp_security.abc
import aiohttp_session
from oslo_config import cfg

from deep_dashboard import config
from deep_dashboard import log

LOG = log.getLogger("deep_dashboard.auth")

iam_opts = [
    cfg.StrOpt(
        'client-id',
        help="""
IAM Client ID to use for authentication
"""),
    cfg.StrOpt(
        'client-secret',
        secret=True,
        help="""
IAM Client Secret corresponding to the configured Client ID
"""),
    cfg.URIOpt(
        'base-url',
        schemes=["http", "https"],
        help="""
IAM Base URL.
"""),
    cfg.URIOpt(
        'authorize-url',
        default="$iam.base_url/authorize",
        schemes=["http", "https"],
        help="""
IAM Authorization endpoint URL.
"""),
    cfg.URIOpt(
        'access-token-url',
        default="$iam.base_url/token",
        schemes=["http", "https"],
        help="""
IAM Access Token endpoint URL.
"""),
    cfg.URIOpt(
        'user-info-url',
        default="$iam.base_url/userinfo",
        schemes=["http", "https"],
        help="""
IAM User Info endpoint URL.
"""),
    cfg.URIOpt(
        'redirect-uri',
        default="https://127.0.0.1/login/iam",
        schemes=["http", "https"],
        help="""
Redirection endpoint. Configure this with your public IP, as configured in
the IAM Client configuration.
"""),
]

CONF = config.CONF
CONF.register_opts(iam_opts, group="iam")


class IamAuthorizationPolicy(aiohttp_security.abc.AbstractAuthorizationPolicy):
    def __init__(self, user_groups=None):
        super().__init__()
        self.user_groups = user_groups

    async def authorized_userid(self, identity):
        """Retrieve authorized user id.
        Return the user_id of the user identified by the identity
        or 'None' if no user exists related to the identity.
        """
        if self.user_groups:
            # Do something with groups FIXME
            pass
        return identity

    async def permits(self, identity, permission, context=None):
        """Check user permissions.
        Return True if the identity is allowed the permission
        in the current context, else return False.
        """
        return True


def get_iam_client():
    client_id = CONF.iam.client_id
    client_secret = CONF.iam.client_secret

    base_url = CONF.iam.base_url
    authorize_url = CONF.iam.authorize_url
    access_token_url = CONF.iam.access_token_url
    user_info_url = CONF.iam.user_info_url
    redirect_url = CONF.iam.redirect_uri

    iam = aioauth_client.OAuth2Client(
        client_id=client_id,
        client_secret=client_secret,
        base_url=base_url,
        authorize_url=authorize_url,
        access_token_url=access_token_url,
        redirect_uri=redirect_url,
        scope='openid email profile offline_access',
    )
    iam.user_info_url = user_info_url

    return iam


async def get_token_userinfo(request, refresh_token=None):
    client = request.app.iam_client

    try:
        if refresh_token is None:
            call = client.get_access_token(request.query)
        else:
            call = client.get_access_token(refresh_token,
                                           grant_type='refresh_token')
        access_token, meta = await call
    except Exception as e:
        exc = web.HTTPInternalServerError(
            reason=f"Could not contact IAM server: {e}"
        )
        raise exc

    if not access_token:
        raise web.HTTPForbidden()

    _, userinfo = await client.user_info()
    delta = datetime.timedelta(seconds=meta["expires_in"])

    # datetime.datetime objects are not JSON serializable
    meta["expires_in"] = (datetime.datetime.now() + delta).timestamp()

    session = await aiohttp_session.get_session(request)
    session["oauth_meta"] = meta

    return meta, userinfo


@web.middleware
async def auth_middleware(request, handler):
    is_authenticated = not await aiohttp_security.is_anonymous(request)
    session = await aiohttp_session.get_session(request)

    request.context["current_user"] = {
        "authenticated": is_authenticated,
    }
    route_name = request.match_info.route.name
    if isinstance(request.match_info.route, web_urldispatcher.ResourceRoute):
        if route_name == "static":
            pass
        elif route_name == "login":
            pass
        elif not is_authenticated and (route_name != "home"):
            session["next"] = str(request.rel_url)
            return web.HTTPFound("/")

    if is_authenticated:

        # Check if user is authorized
        await aiohttp_security.check_authorized(request)

        # Renew the token if its life is too short
        now = datetime.datetime.now()
        # expires_in is a POSIX timestamp
        expires = session["oauth_meta"]["expires_in"]
        expires = datetime.datetime.fromtimestamp(expires)
        delta = datetime.timedelta(seconds=20 * 60)
        if (expires - now) < delta:
            refresh_token = session["oauth_meta"].get("refresh_token")
            try:
                meta, userinfo = await get_token_userinfo(
                    request,
                    refresh_token=refresh_token
                )
            except web.HTTPForbidden:
                session["next"] = str(request.rel_url)
                return web.HTTPFound("/")

        request.context["current_user"]["username"] = session["username"]
        request.context["current_user"]["gravatar"] = session["gravatar"]

    response = await handler(request)
    return response



# -*- coding: utf-8 -*-

# Copyright 2019 Spanish National Research Council (CSIC)
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import pathlib

from oslo_config import cfg
from oslo_log import log

from deep_dashboard import version

opts = [
    cfg.StrOpt(
        "github-secret",
        secret=True,
        help="""
GitHub secret to trigger reloading of modules
"""),
    cfg.URIOpt(
        'deep-oc-modules',
        default=None,
        schemes=["http", "https"],
        deprecated_for_removal=True,
        help="""
URL of the DEEP OC modules YAML file. This option is marked for removal.
"""),
    cfg.URIOpt(
        'deep-oc-repo',
        default="https://github.com/deephdc/deep-oc/",
        schemes=["http", "https"],
        help="""
URL of the DEEP OC repository.
"""),
    cfg.StrOpt(
        'deep-oc-dir',
        default="$runtime_dir/deep-oc",
        help="""
Path to the directory where to store all the DEEP OC modules.
"""),
    cfg.BoolOpt(
        "purge-deep-oc-directory",
        default=False,
        help="""
Remove $deep_oc_dir in case it is not a valid Git repository.
"""),
    cfg.StrOpt(
        "runtime-dir",
        default="/var/run/deep-dashboard",
        help="""
Define the DEEP dashboard base runtime directory
"""),
    cfg.StrOpt(
        "static-path",
        default=(pathlib.Path(__file__).parent / "static").as_posix(),
        help="""
Path where the static files are stored.
"""),
]


orchestrator_opts = [
    cfg.URIOpt(
        'url',
        required=True,
        schemes=["http", "https"],
        help="""
DEEP orchestrator endpoint.
"""),
    cfg.StrOpt(
        'tosca-dir',
        default="$runtime_dir/tosca-templates",
        help="""
Path to the directory where to store all the orchestrator TOSCA templates.
"""),
    cfg.StrOpt(
        'deep-templates-dir',
        default="deep-oc",
        help="""
Directory relative to $tosca_dir where the DEEP TOSCA templates are stored.
"""),
    cfg.BoolOpt(
        "purge-tosca-directory",
        default=False,
        help="""
Remove $tosca_dir in case it is not a valid Git repository.
"""),
    cfg.StrOpt(
        'tosca-parameters-dir',
        default="$runtime_dir/tosca-parameters-dir",
        help="""
Path to the directory where the additional TOSCA parameters are stored.
"""),
    cfg.URIOpt(
        'tosca-repo',
        default="https://github.com/indigo-dc/tosca-templates/",
        schemes=["http", "https", "git"],
        help="""
URL of the DEEP tosca templates repository.
"""),
    cfg.DictOpt(
        "common-toscas",
        default={
            "default (with remote storage)": "deep-oc-marathon-webdav.yml",
            "default (without remote storage)": "deep-oc-marathon-minimal.yml",
        },
    ),
]

cache_opts = [
    cfg.StrOpt(
        "memcached-ip",
        help="""
IP of the memcached server to use.

If not set, we will not use memcached at all, therefore the DEEP dashboard
will not behave as expected when using several workers.
"""),
    cfg.PortOpt(
        "memcached-port",
        default=11211,
        help="""
Port of the memcached server to use.
"""),
]

cli_opts = [
    cfg.StrOpt('listen-ip',
               help="""
IP address on which the DEEP Dashboard will listen.

The DEEP dashboard service will listen on this IP address.
"""),
    cfg.PortOpt('listen-port',
                help="""
Port on which the DEEP Dashboard will liste,.

The DEEP dashboard service will listen on this port.
"""),
    cfg.StrOpt('listen-path',
               help="""
Path to the UNIX socket where the DEEP dashboard will listen.
"""),
]

CONF = cfg.CONF
CONF.register_cli_opts(cli_opts)
CONF.register_opts(opts)
CONF.register_opts(orchestrator_opts, group="orchestrator")
CONF.register_opts(cache_opts, group="cache")


def parse_args(args, default_config_files=None):
    cfg.CONF(args,
             project='deep_dashboard',
             version=version.__version__,
             default_config_files=default_config_files)


def prepare_logging():
    log.register_options(CONF)
    log.set_defaults(default_log_levels=log.get_default_log_levels())


def configure(argv, default_config_files=None):
    prepare_logging()
    parse_args(argv, default_config_files=default_config_files)
    log.setup(CONF, "deep_dashboard")



# -*- coding: utf-8 -*-

# Copyright 2019 Spanish National Research Council (CSIC)
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import base64
import collections
import concurrent.futures
import pathlib

import aiocache
from aiohttp import web
import aiohttp_jinja2
import aiohttp_security
import aiohttp_session
import aiohttp_session.cookie_storage
import aiohttp_session.memcached_storage
import aiohttp_session_flash
import aiojobs
import aiomcache
from cryptography import fernet
import jinja2
from oslo_concurrency import lockutils

from deep_dashboard import auth
from deep_dashboard import config
from deep_dashboard import deep_oc
from deep_dashboard.handlers import base
from deep_dashboard.handlers import deployments
from deep_dashboard.handlers import modules
from deep_dashboard import log
from deep_dashboard import version

__version__ = version.__version__

CONF = config.CONF
LOG = log.LOG


@web.middleware
async def error_middleware(request, handler):
    try:
        response = await handler(request)
        if response.status != 404:
            return response
        message = response.message
    except web.HTTPServerError as e:
        LOG.exception(e)
        message = "Internal server error. "
    except web.HTTPException as e:
        message = f"Error {e.status_code}: {e.reason}"
    aiohttp_session_flash.flash(request, ("danger", message))
    response = web.HTTPFound("/")
    return response


@web.middleware
async def meta_middleware(request, handler):
    request.context = {
        "meta": {
            "version": __version__,
        }
    }
    response = await handler(request)
    return response


class CacheManager:
    def __init__(self, cache, namespace):
        self.cache = cache
        self.namespace = namespace

    async def set_values(self, values):
        if isinstance(values, dict):
            values = values.items()
        await self.cache.multi_set(values, namespace=self.namespace)
        await self.cache.set("__all__", [i[0] for i in values],
                             namespace=self.namespace)

    async def get(self, key):
        return await self.cache.get(key, namespace=self.namespace)

    async def get_all_keys(self):
        result = await self.cache.get("__all__", namespace=self.namespace)
        return result or []

    async def get_all(self):
        keys = await self.cache.get("__all__", namespace=self.namespace)
        if not keys:
            return {}
        values = await self.cache.multi_get(keys, namespace=self.namespace)
        return dict(zip(keys, values))

    async def exists(self, key):
        return await self.cache.exists(key, namespace=self.namespace)


async def init(args):
    LOG.info("Starting DEEP Dashboard...")

    runtime_dir = pathlib.Path(CONF.runtime_dir)

    runtime_dir.mkdir(parents=True, exist_ok=True)

    app = web.Application(debug=True)
    app.runtime_dir = runtime_dir
    lockutils.set_defaults(runtime_dir)

    tpl_path = pathlib.Path(__file__).parent / "templates"
    aiohttp_jinja2.setup(
        app,
        context_processors=[aiohttp_session_flash.context_processor],
        loader=jinja2.FileSystemLoader(tpl_path)
    )

    app.iam_client = auth.get_iam_client()

    base.routes.static('/static', CONF.static_path, name="static")
    app.add_routes(base.routes)
    app.add_routes(deployments.routes)
    app.add_routes(modules.routes)

    if CONF.cache.memcached_ip:
        mc = aiomcache.Client(CONF.cache.memcached_ip,
                              CONF.cache.memcached_port)
        sess_storage = aiohttp_session.memcached_storage.MemcachedStorage(
            mc,
            cookie_name='DEEPDASHBOARD_M'
        )
        aiocache.caches.add('default', {
            'cache': "aiocache.MemcachedCache",
            'endpoint': CONF.cache.memcached_ip,
            'port': CONF.cache.memcached_port,
        })
    else:
        LOG.warning("Not using memcached, unexpected behaviour when running "
                    "more than one worker!")

        # secret_key must be 32 url-safe base64-encoded bytes
        fernet_key = fernet.Fernet.generate_key()

        secret_key = base64.urlsafe_b64decode(fernet_key)

        sess_storage = aiohttp_session.cookie_storage.EncryptedCookieStorage(
            secret_key,
            cookie_name='DEEPDASHBOARD_E'
        )
        aiocache.caches.add('default', {
            'cache': "aiocache.SimpleMemoryCache",
        })

    aiohttp_session.setup(app, sess_storage)

    policy = aiohttp_security.SessionIdentityPolicy()
    aiohttp_security.setup(app, policy, auth.IamAuthorizationPolicy())

    app.middlewares.append(meta_middleware)
    app.middlewares.append(aiohttp_session_flash.middleware)
    app.middlewares.append(auth.auth_middleware)
    app.middlewares.append(error_middleware)
    app.modules = {}
    cache = aiocache.caches.get("default")
    app.cache = collections.namedtuple(
        "Cache",
        ["modules", "tosca_templates"],
        defaults=[CacheManager(cache, "modules"),
                  CacheManager(cache, "tosca")]
    )()

    app.scheduler = await aiojobs.create_scheduler()
    app.pool = concurrent.futures.ThreadPoolExecutor()
    app.on_startup.append(deep_oc.download_catalog)
    app.on_startup.append(deep_oc.load_catalog)

    return app



# -*- coding: utf-8 -*-

# Copyright 2021 Spanish National Research Council (CSIC)
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import asyncio
import json

import aiohttp
import aiohttp_session

from deep_dashboard import config
from deep_dashboard import orchestrator

CONF = config.CONF


def is_enabling_services(deployment_type, service_type):
    if deployment_type == "CLOUD":
        return service_type in ["org.openstack.nova", "com.amazonaws.ec2"]
    elif deployment_type == "MARATHON":
        return service_type in ["eu.indigo-datacloud.marathon"]
    elif deployment_type == "CHRONOS":
        return service_type in ["eu.indigo-datacloud.chronos"]
    elif deployment_type == "QCG":
        return service_type in ["eu.deep.qcg"]
    else:
        return True


async def get_slas(request, deployment_type=""):
    cli = await orchestrator.get_client(
        CONF.orchestrator.url,
        request
    )
    orchestrator_config = cli.config.get()

    slam_url = orchestrator_config.slam_url
    cmdb_url = orchestrator_config.cmdb_url
    access_token = request.app.iam_client.access_token

    headers = {
        'Authorization': f'bearer {access_token}'
    }

    session = await aiohttp_session.get_session(request)
    url = slam_url + "/preferences/" + session['organisation_name']

    cli_session = aiohttp.ClientSession()
    async with cli_session.get(url,
                               headers=headers,
                               raise_for_status=True) as r:
        content = await r.json()

    slas = content['sla']

    filtered_slas = []
    for sla in slas:
        service_id = sla['services'][0]['service_id']
        url = cmdb_url + "/service/id/" + service_id
        async with cli_session.get(url,
                                   headers=headers,
                                   raise_for_status=True) as r:
            # Unfortunately we cannot use r.json() here as the CMDB does not
            # send a proper content-type for the file, as it is being sent
            # as plaintext
            content = await r.read()
        content = json.loads(content)

        data = content["data"]
        service_type = data['service_type']
        sitename = data['sitename']
        endpoint = data.get('endpoint')
        iam_enabled = data.get('iam_enabled')

        if iam_enabled is None:
            iam_enabled = "True"

        gpu_support = data.get("properties", {}).get("gpu_support", False)
        if gpu_support:
            service_type = f"{service_type} (gpu_support: {gpu_support})"

        if is_enabling_services(deployment_type, service_type):
            sla['service_id'] = sla['services'][0]['service_id']
            sla['service_type'] = service_type
            sla['sitename'] = sitename
            sla['endpoint'] = endpoint
            sla['iam_enabled'] = iam_enabled

            filtered_slas.append(sla)

    return filtered_slas


async def load_slas(request):
    if not hasattr(request.app, "slas"):
        request.app.slas = []
    slas = await get_slas(request)
    request.app.slas = slas


async def load_slas_as_task(request):
    return asyncio.create_task(load_slas(request))



# -*- coding: utf-8 -*-

# Copyright 2019 Spanish National Research Council (CSIC)
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import aiohttp
from aiohttp import web
import aiohttp_jinja2
import aiohttp_session_flash as flash
import orpy.exceptions
import yaml

from deep_dashboard import config
from deep_dashboard import log
from deep_dashboard import orchestrator

CONF = config.CONF
LOG = log.LOG

CONF = config.CONF

routes = web.RouteTableDef()


@routes.get("/deployments", name="deployments")
@aiohttp_jinja2.template('deployments/index.html')
async def get_deployments(request):
    request.context["deployments"] = []

    cli = await orchestrator.get_client(
        CONF.orchestrator.url,
        request
    )

    request.context["breadcrumbs"] = [
        ("Home", False, "/"),
        ("Deployments", True, "/deployments"),  # FIXME(aloga): use url
    ]

    try:
        request.context["deployments"] = cli.deployments.list()

        # Make creation times readable:
        # 2022-10-21T09:22+0000 --> 2022-10-21 09:22
        for dep in request.context["deployments"]:
            day, hour = dep.creationTime.split('T')
            dep.creationTime = f'{day} 🕑 {hour[:5]}'

            day, hour = dep.updateTime.split('T')
            dep.updateTime = f'{day} 🕑 {hour[:5]}'

    except orpy.exceptions.ClientException as e:
        flash.flash(
            request,
            ("danger", f"Error retrieving deployment list: {e.message}"),
        )
    finally:
        return request.context


@routes.get("/deployments/{uuid}/template", name="deployments.template")
@aiohttp_jinja2.template('deployments/template.html')
async def get_deployment_template(request):
    request.context["deployments"] = []

    uuid = request.match_info['uuid']

    request.context["breadcrumbs"] = [
        ("Home", False, "/"),
        ("Deployments", False, "/deployments"),  # FIXME(aloga): use url
        (uuid, False, f"/deployments/{uuid}"),  # FIXME(aloga): use url
        ("template", True, f"/deployments/{uuid}/template"),  # FIXME(aloga)
    ]

    cli = await orchestrator.get_client(
        CONF.orchestrator.url,
        request
    )

    try:
        template = cli.deployments.get_template(uuid).template
        request.context["template"] = template
    except orpy.exceptions.ClientException as e:
        flash.flash(
            request,
            ("danger", f"Error retrieving deployment {uuid} template: "
                       f"{e.message}"),
        )
        return web.HTTPFound("/deployments")
    else:
        return request.context


# FIXME(aloga): this is not correct, we should not use a GET but a DELETE
@routes.get("/deployments/{uuid}/delete", name="deployments.delete")
async def delete_deployment(request):
    uuid = request.match_info['uuid']

    cli = await orchestrator.get_client(
        CONF.orchestrator.url,
        request
    )

    try:
        cli.deployments.delete(uuid)
    except orpy.exceptions.ClientException as e:
        flash.flash(
            request,
            ("danger", f"Error deleting deployment {uuid}: {e.message}")
        )
    finally:
        return web.HTTPFound("/deployments")


# FIXME(aloga): this is not correct, we should not use a GET but a DELETE
@routes.get("/deployments/{uuid}/history/{model}/{training_uuid}/delete",
            name="training.delete")
async def delete_training(request):
    dep_uuid = request.match_info["uuid"]
    model = request.match_info["model"]
    training_uuid = request.match_info["training_uuid"]

    cli = await orchestrator.get_client(
        CONF.orchestrator.url,
        request
    )

    try:
        deployment = cli.deployments.show(dep_uuid)
    except orpy.exceptions.ClientException as e:
        flash.flash(
            request,
            ("danger", f'Error getting deployment {dep_uuid}: {e.message}')
        )
        return web.HTTPFound("/deployments")

    # Check if deployment is still in 'create_in_progress'
    if 'deepaas_endpoint' not in deployment.outputs:
        flash.flash(
            request,
            ("warning",
             'Wait until creation is completed before you access the '
             'training history.')
        )
        return web.HTTPFound(f"/deployments/{dep_uuid}/history")

    # Check if deployment has DEEPaaS V2
    deepaas_url = deployment.outputs['deepaas_endpoint']
    training_url = f"{deepaas_url}/v2/models/{model}/train/{training_uuid}"

    session = aiohttp.ClientSession()

    try:
        async with session.delete(training_url,
                                  raise_for_status=True) as r:
            if r.status not in [200, 201]:
                raise Exception
    except Exception as e:
        flash.flash(
            request,
            ("warning", f"Could not delete training!! (reason: {e.message})")
        )
    finally:
        return web.HTTPFound(f"/deployments/{dep_uuid}/history")


@routes.get("/deployments/{uuid}/history", name="deployments.history")
@aiohttp_jinja2.template('deployments/summary.html')
async def show_deployment_history(request):
    uuid = request.match_info['uuid']

    request.context["breadcrumbs"] = [
        ("Home", False, "/"),
        ("Deployments", False, "/deployments"),  # FIXME(aloga): use url
        (uuid, False, f"/deployments/{uuid}"),  # FIXME(aloga): use url
        ("history", True, f"/deployments/{uuid}/history"),  # FIXME(aloga)
    ]

    cli = await orchestrator.get_client(
        CONF.orchestrator.url,
        request
    )

    try:
        deployment = cli.deployments.show(uuid)
    except orpy.exceptions.ClientException as e:
        flash.flash(
            request,
            ("danger", f'Error getting deployment {uuid}: {e.message}')
        )
        return web.HTTPFound("/deployments")

    # Check if deployment is still in 'create_in_progress'
    if 'deepaas_endpoint' not in deployment.outputs:
        flash.flash(
            request,
            ("warning",
             'Wait until creation is completed before you access the '
             'training history.')
        )
        return web.HTTPFound("/deployments")

    session = aiohttp.ClientSession()

    # Check if deployment has DEEPaaS V2
    deepaas_url = deployment.outputs['deepaas_endpoint']
    try:
        async with session.get(deepaas_url,
                               raise_for_status=True) as r:
            data = await r.json()
        versions = data['versions']
        if 'v2' not in [v['id'] for v in versions]:
            raise Exception
    except Exception:
        flash.flash(
            request,
            ("warning",
             "You need to be running DEEPaaS V2 inside the deployment "
             "to be able to access the training history.")
        )
        return web.HTTPFound("/deployments")

    # Get info
    async with session.get(deepaas_url + '/v2/models',
                           raise_for_status=False) as r:
        r = await r.json()

    training_info = {}
    for model in r['models']:
        async with session.get(f'{deepaas_url}/v2/models/{model["id"]}/train/',
                               raise_for_status=False) as r:
            training_info[model['id']] = await r.json()

    request.context["deployment"] = deployment
    request.context["training_info"] = training_info
    return request.context


@routes.post("/deployments")
async def create_module_training(request):
    form_data = await request.post()
    LOG.debug(f"Received form data: {form_data}")

    tosca_template = form_data.get('template')
    LOG.debug(f"Selected template file: {tosca_template}")

    if not await request.app.cache.tosca_templates.exists(tosca_template):
        flash.flash(
            request,
            ("danger", f"TOSCA template does not exist: {tosca_template}.")
        )
        return web.HTTPFound("/modules")

    aux = await request.app.cache.tosca_templates.get(tosca_template)
    template = aux["original tosca"]

    params = {}
    if 'extra_opts.keepLastAttempt' in form_data:
        params["keepLastAttempt"] = "true"
    else:
        params["keepLastAttempt"] = "false"

    if form_data['extra_opts.schedtype'] == "manual":
        if sla := form_data.get('extra_opts.selectedSLA', None):
            LOG.debug(f"Adding SLA to deployment request {sla}")
            template['topology_template']['policies'] = [{
                "deploy_on_specific_site": {
                    "type": "tosca.policies.indigo.SlaPlacement",
                    "properties": {
                        "sla_id": sla
                    }
                }
            }]
        else:
            flash.flash(
                request,
                ("danger", "SLA does not exist.")
            )

    inputs = {k: v for (k, v) in form_data.items()
              if not k.startswith("extra_opts.")}
    LOG.debug("Parameters:", inputs)

    template = yaml.dump(template,
                         default_flow_style=False,
                         sort_keys=False)
    LOG.debug(f"Final template generated {template}")

    # FIXME(aloga): Make this async
    cli = await orchestrator.get_client(
        CONF.orchestrator.url,
        request
    )

    cli.deployments.create(template, parameters=inputs)
    return web.HTTPFound("/deployments")



# -*- coding: utf-8 -*-

# Copyright 2021 Spanish National Research Council (CSIC)
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import asyncio
import collections
import json
import os.path
import pathlib
import shutil
import subprocess

import aiohttp
import git
import git.exc
from oslo_concurrency import lockutils

from deep_dashboard import config
from deep_dashboard import log
from deep_dashboard import tosca

CONF = config.CONF
LOG = log.getLogger("deep_dashboard.deep_oc")


@lockutils.synchronized("catalog.lock", external=True)
def download_deep_catalog():
    """Get and load modules in the DEEP marketplace as TOSCA files."""

    # FIXME(aloga): when using multiple backends this funcion is executed in
    # all of them on startup. This does not cause errors (it will try to
    # update git git repo, but this will delay startup time. We need to use
    # fasteners to get the lock and return if the repo is already locked.
    LOG.debug(f"Downloading DEEP OC from {CONF.deep_oc_repo}")

    deep_oc_dir = pathlib.Path(CONF.deep_oc_dir)

    try:
        git.Repo(deep_oc_dir)
    except git.exc.NoSuchPathError:
        deep_oc_dir.mkdir(parents=True)
        git.Repo.clone_from(CONF.deep_oc_repo, deep_oc_dir)
    except (git.exc.InvalidGitRepositoryError, git.exc.GitCommandError):
        if CONF.purge_deep_oc_directory:
            shutil.rmtree(deep_oc_dir)
            deep_oc_dir.mkdir(parents=True)
            git.Repo.clone_from(CONF.deep_oc_repo, deep_oc_dir)
        else:
            raise

    g = git.cmd.Git(deep_oc_dir)
    g.pull()
    # NOTE(aloga): we cannot rely on gitpython to update the submodules, as
    # it fails to update them if the branch is other than main
    subprocess.run(["git", "submodule", "init"],
                   cwd=deep_oc_dir)
    subprocess.run(["git", "submodule", "update", "--recursive"],
                   cwd=deep_oc_dir)


@lockutils.synchronized("catalog.lock", external=True)
async def load_modules_metadata():
    modules_meta = collections.OrderedDict()

    LOG.debug(f"Loading DEEP OC from {CONF.deep_oc_repo}")
    deep_oc_dir = pathlib.Path(CONF.deep_oc_dir)
    repo = git.Repo(deep_oc_dir)

    for sm in repo.submodules:
        meta_file = deep_oc_dir / deep_oc_dir / sm.name / "metadata.json"
        try:
            with open(meta_file, "r") as f:
                metadata = f.read()
            metadata = json.loads(metadata)
        except FileNotFoundError:
            LOG.error(f"Cannot find metadata file {meta_file}")
        except json.JSONDecodeError:
            LOG.error(f"Cannot decode JSON file {meta_file}")
            continue
        metadata["module_url"] = sm.url
        name = sm.name.lower().replace('_', '-')

        modules_meta[name] = metadata
        modules_meta[name]["original_name"] = sm.name
        modules_meta[name]["deepaas_name"] = sm.name.replace("DEEP-OC-", "", 1)

    modules_meta["external"] = {
        'title': 'Run your own module',
        'summary': 'Use your own external container hosted '
                   'in Dockerhub',
        'tosca': [],
        'keywords': [],
        'docker_tags': ['latest'],
        'sources': {
            'docker_registry_repo': ''
        },
        'module_url': "",
    }

    return modules_meta


async def get_dockerhub_tags(session, image):
    url = f'https://registry.hub.docker.com/v2/repositories/{image}/tags'
    try:
        async with session.get(url, raise_for_status=True) as r:
            aux = await r.json()
    except aiohttp.client_exceptions.ClientError:
        LOG.error(f"Cannot get tags from DockerHub {image}")
        return []
    return [i['name'] for i in aux["results"]]


@lockutils.synchronized("tosca-templates.lock", external=True)
@lockutils.synchronized("catalog.lock", external=True)
async def map_modules_to_tosca(modules_metadata, tosca_templates):
    session = aiohttp.ClientSession()

    tosca_dir = pathlib.Path(CONF.orchestrator.tosca_dir)
    common_toscas = CONF.orchestrator.common_toscas

    for module_name, metadata in modules_metadata.items():
        toscas = collections.OrderedDict()
        # Find tosca names from metadata
        for t in metadata['tosca']:
            tosca_name = os.path.basename(t['url'])
            if tosca_name in common_toscas.values():
                continue
            try:
                if tosca_name not in tosca_templates:
                    tosca_file = tosca_dir / tosca_name
                    async with session.get(t["url"],
                                           raise_for_status=True) as r:
                        with open(tosca_file, "w") as f:
                            f.write(await r.data())
                toscas[t['title'].lower()] = tosca_name
            except Exception as e:
                LOG.warning(f'Error processing TOSCA in module '
                            f'{module_name} from {t["url"]}: {e}')

        # Add always common TOSCAs
        for k, v in common_toscas.items():
            toscas[k] = v

        # Add Docker tags
        if metadata['sources']['docker_registry_repo']:
            dockerhub_tags = await get_dockerhub_tags(
                session,
                metadata['sources']['docker_registry_repo']
            )
            metadata.setdefault('docker_tags', [])
            if metadata['docker_tags']:
                # Check that the tags provided by the user are indeed
                # present in DockerHub
                aux = set(metadata['docker_tags'])
                aux = aux.intersection(set(dockerhub_tags))
                aux = sorted(list(aux))
                metadata['docker_tags'] = aux
            else:
                metadata['docker_tags'] = dockerhub_tags

        metadata["tosca_templates"] = toscas

    await session.close()
    return modules_metadata


async def download_catalog(app):
    loop = asyncio.get_event_loop()
    app.tosca_downloader = loop.run_in_executor(app.pool,
                                                tosca.download_tosca_templates)
    app.catalog_downloader = loop.run_in_executor(app.pool,
                                                  download_deep_catalog)


async def load_catalog(app):
    await app.scheduler.spawn(_load_catalog(app))


async def _load_catalog(app):
    async def inner():
        while True:
            if not all([app.tosca_downloader.done(),
                        app.catalog_downloader.done()]):
                await asyncio.sleep(5)
            else:
                break

        tosca_templates, modules_meta = await asyncio.gather(
            asyncio.create_task(tosca.load_tosca_templates()),
            asyncio.create_task(load_modules_metadata())
        )
        modules_meta = await map_modules_to_tosca(modules_meta,
                                                  tosca_templates)
        app.modules = modules_meta
        app.tosca_templates = tosca_templates
        await app.cache.modules.set_values(modules_meta)
        await app.cache.tosca_templates.set_values(tosca_templates)
    await inner()



# -*- coding: utf-8 -*-

# Copyright 2021 Spanish National Research Council (CSIC)
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import copy
import pathlib
import shutil
import yaml

import git
import git.cmd
import git.exc
from oslo_concurrency import lockutils

from deep_dashboard import config
from deep_dashboard import log

CONF = config.CONF
LOG = log.getLogger("deep_dashboard.tosca")

tosca_info_defaults = {
    "valid": False,
    "description": "TOSCA Template",
    "metadata": {
        "icon": "https://cdn4.iconfinder.com/data/icons/"
        "mosaicon-04/512/websettings-512.png"
    },
    "enable_config_form": False,
    "inputs": {},
    "tabs": {},
}


@lockutils.synchronized("tosca-templates.lock", external=True)
def download_tosca_templates():
    """Load DEEP-OC related TOSCA templates from configured repository."""

    # FIXME(aloga): when using multiple backends this funcion is executed in
    # all of them on startup. This does not cause errors (it will try to
    # update git git repo, but this will delay startup time. We need to use
    # fasteners to get the lock and return if the repo is already locked.

    LOG.debug("Downloading TOSCA templates")

    tosca_dir = pathlib.Path(CONF.orchestrator.tosca_dir)

    try:
        git.Repo(tosca_dir)
    except git.exc.NoSuchPathError:
        tosca_dir.mkdir(parents=True)
        git.Repo.clone_from(CONF.orchestrator.tosca_repo, tosca_dir)
    except (git.exc.InvalidGitRepositoryError, git.exc.GitCommandError):
        if CONF.orchestrator.purge_tosca_directory:
            shutil.rmtree(tosca_dir)
            tosca_dir.mkdir(parents=True)
            git.Repo.clone_from(CONF.orchestrator.tosca_repo, tosca_dir)
        else:
            raise

    g = git.cmd.Git(tosca_dir)
    g.pull()


@lockutils.synchronized("tosca-templates.lock", external=True)
async def load_tosca_templates():
    LOG.debug("Loading TOSCA templates")

    tosca_dir = pathlib.Path(CONF.orchestrator.tosca_dir)

    toscas = {}
    deep_tpl_dir = tosca_dir / CONF.orchestrator.deep_templates_dir
    for tosca_file in deep_tpl_dir.rglob("*y*ml"):
        with open(tosca_file, "r") as tpl:
            tpl = yaml.full_load(tpl)

        if not tpl:
            continue

        tosca_info = copy.deepcopy(tosca_info_defaults)
        if 'topology_template' not in tpl:
            continue

        tosca_info["valid"] = True
        tosca_info["original tosca"] = copy.deepcopy(tpl)

        # Update meta with sub dict
        meta = tpl.pop("metadata", {})
        tosca_info["metadata"].update(meta)

        # Update description if it exists
        tosca_info["description"] = tpl.get("description",
                                            tosca_info["description"])

        # Update deployment type
        topology = tpl.pop("topology_template", {})
        node_templates = topology.get("node_templates", {})
        tosca_info['deployment_type'] = get_deployment_type(node_templates)

        # Update inputs
        inputs = topology.pop("inputs", {})
        tosca_info["inputs"] = inputs

        # Add parameters code here
        if CONF.orchestrator.tosca_parameters_dir:
            prm_dir = pathlib.Path(CONF.orchestrator.tosca_parameters_dir)
            tosca_name = tosca_file.stem

            for f in prm_dir.glob(f"{tosca_name}.parameters.y*ml"):
                with open(f) as f:
                    params = yaml.full_load(f)
                if not params:
                    continue

                tosca_info['enable_config_form'] = True
                tosca_info['inputs'].update(params.get("inputs", {}))
                tosca_info['tabs'].update(params.get("tabs", {}))

        toscas[tosca_file.name] = tosca_info

    return toscas


def get_deployment_type(nodes):
    deployment_type = ""
    type_map = {
        "tosca.nodes.indigo.Compute": "CLOUD",
        "tosca.nodes.indigo.Container.Application.Docker.Marathon": "MARATHON",
        "tosca.nodes.indigo.Container.Application.Docker.Chronos": "CHRONOS",
        "tosca.nodes.indigo.Qcg.Job": "QCG",
    }
    for (j, u) in nodes.items():
        for (k, v) in u.items():
            if k == "type":
                deployment_type = type_map.get(v, "")
            if deployment_type:
                return deployment_type
    return deployment_type
